'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var

  // Mising "types"
  //  - Comment
  //  - BlockComment
  //  - &attributes
  //  - InterpolatedTag
  //  - Code(buffer: false + block)
  //  - Each(object)

  babel = _ref.babel;
  var parse = _ref.parse;
  var helpers = _ref.helpers;
  var ast = _ref.ast;
  var path = _ref.path;
  var code = _ref.code;
  var types = babel.types;

  var baseLine = path.node.loc.start.line;
  var lastLine = 0;
  var t = {};
  function getFn(key) {
    return function () {
      var res = types[key].apply(types, arguments);
      var loc = { line: baseLine + lastLine, column: 0 };
      res.loc = { start: loc, end: loc };
      return res;
    };
  }
  for (var key in types) {
    if (/^is|^assert/.test(key)) {
      t[key] = types[key];
    } else {
      t[key] = getFn(key);
    }
  }
  var nodes = t.identifier('pug_nodes');
  function withString(node, stringLiteral) {
    t.assertStringLiteral(stringLiteral);
    if (t.isStringLiteral(node)) {
      return t.stringLiteral(node.value + stringLiteral.value);
    }
    if (t.isBinaryExpression(node, { operator: '+' }) && t.isStringLiteral(node.right)) {
      return t.binaryExpression('+', node.left, withString(node.right, stringLiteral));
    }
    return t.binaryExpression('+', node, stringLiteral);
  }
  var staticBlockID = 0;
  var compiler = {
    baseBlock: function baseBlock() {
      return {
        _id: 'base',
        getKey: function getKey(fn) {
          fn(t.stringLiteral('pug'));
        },
        handleAttributes: function handleAttributes(attrs) {},
        end: function end() {}
      };
    },
    staticBlock: function staticBlock(parent) {
      var enabled = false;
      var parentEnabled = false;
      var key = null;
      var pending = [];
      var index = 0;
      parent.getKey(function (parentKey) {
        parentEnabled = true;
        key = withString(parentKey, t.stringLiteral(':' + staticBlockID++));
        onUpdate();
      });
      function onUpdate() {
        if (enabled && parentEnabled) {
          while (pending.length) {
            pending.shift()(withString(key, t.stringLiteral(':' + index++)));
          }
        }
      }
      return {
        getKey: function getKey(fn) {
          if (pending.indexOf(fn) === -1) {
            pending.push(fn);
          }
          onUpdate();
        },
        handleAttributes: function handleAttributes(attrs) {
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
            for (var _iterator = attrs[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              var attr = _step.value;

              if (t.isJSXAttribute(attr) && t.isJSXIdentifier(attr.name, { name: 'key' })) {
                return;
              }
            }
          } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
              }
            } finally {
              if (_didIteratorError) {
                throw _iteratorError;
              }
            }
          }

          this.getKey(function (key) {
            if (t.isStringLiteral(key)) key = key; // t.jSXText(key.value);
            else key = t.jSXExpressionContainer(key);
            attrs.push(t.jSXAttribute(t.jSXIdentifier('key'), key));
          });
        },
        end: function end() {
          enabled = true;
          onUpdate();
        }
      };
    },
    dynamicBlock: function dynamicBlock(parent, lineNumber) {
      var enabled = false;
      var localKey = null;
      var parentEnabled = false;
      var parentKey = null;
      var pending = [];
      var index = 0;
      function join(left, right) {
        return t.binaryExpression('+', left, right);
      }
      parent.getKey(function (_parentKey) {
        parentEnabled = true;
        parentKey = withString(_parentKey, t.stringLiteral(':'));
        onUpdate();
      });
      function onUpdate() {
        if (enabled && parentEnabled && localKey) {
          while (pending.length) {
            pending.shift()(withString(join(parentKey, localKey), t.stringLiteral(':' + index++)));
          }
        } else if (enabled && parentEnabled) {
          var err = (0, _pugError2.default)('MISSING_KEY', 'You must specify a key for the first item in any loops.', {
            line: path.node.loc.start.line + lineNumber,
            filename: 'pug',
            src: code
          });
          throw err;
        }
      }
      return {
        getKey: function getKey(fn) {
          if (pending.indexOf(fn) === -1) {
            pending.push(fn);
          }
          onUpdate();
        },
        handleAttributes: function handleAttributes(attrs) {
          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            for (var _iterator2 = attrs[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              var attr = _step2.value;

              if (t.isJSXAttribute(attr) && t.isJSXIdentifier(attr.name, { name: 'key' })) {
                if (localKey === null && t.isJSXExpressionContainer(attr.value) && attr.value.expression) {
                  localKey = attr.value.expression;
                  onUpdate();
                  // remove the attribute and replace with the properly nested version
                  attrs.splice(attrs.indexOf(attr), 1);
                } else {
                  return;
                }
              }
            }
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }

          this.getKey(function (key) {
            if (t.isStringLiteral(key)) key = key; // t.jSXText(key.value);
            else key = t.jSXExpressionContainer(key);
            attrs.push(t.jSXAttribute(t.jSXIdentifier('key'), key));
          });
        },
        end: function end() {
          enabled = true;
          onUpdate();
        }
      };
    },
    parseExpression: function parseExpression(src) {
      var val = parse('x = (' + src + ');').program.body;
      (0, _assert2.default)(val.length === 1);
      return val[0].expression.right;
    },
    parseStatement: function parseStatement(src) {
      var val = parse(src).program.body;
      (0, _assert2.default)(val.length === 1);
      return val[0];
    },
    wrapExpressionInJSX: function wrapExpressionInJSX(value) {
      // returns ["JSX"]
      if (t.isJSX(value)) {
        return value;
      } else if (t.isStringLiteral(value)) {
        return t.jSXText(value.value);
      } else {
        return t.jSXExpressionContainer(value);
      }
    },
    joinJsExpressions: function joinJsExpressions(values) {
      var vals = [];
      // flatten one level
      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;
      var _iteratorError3 = undefined;

      try {
        for (var _iterator3 = values[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
          var val = _step3.value;

          if (t.isArrayExpression(val)) {
            var _iteratorNormalCompletion4 = true;
            var _didIteratorError4 = false;
            var _iteratorError4 = undefined;

            try {
              for (var _iterator4 = val.entries[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
                var e = _step4.value;

                vals.push(e);
              }
            } catch (err) {
              _didIteratorError4 = true;
              _iteratorError4 = err;
            } finally {
              try {
                if (!_iteratorNormalCompletion4 && _iterator4.return) {
                  _iterator4.return();
                }
              } finally {
                if (_didIteratorError4) {
                  throw _iteratorError4;
                }
              }
            }
          } else {
            vals.push(val);
          }
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return) {
            _iterator3.return();
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }

      if (vals.length === 0) {
        return t.nullLiteral();
      } else if (vals.length === 1) {
        return vals[0];
      } else {
        return t.arrayExpression(vals);
      }
    },
    joinJsStatements: function joinJsStatements(values) {
      var vals = [];
      // flatten one level
      var _iteratorNormalCompletion5 = true;
      var _didIteratorError5 = false;
      var _iteratorError5 = undefined;

      try {
        for (var _iterator5 = values[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
          var val = _step5.value;

          if (Array.isArray(val)) {
            var _iteratorNormalCompletion6 = true;
            var _didIteratorError6 = false;
            var _iteratorError6 = undefined;

            try {
              for (var _iterator6 = val[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
                var e = _step6.value;

                vals.push(e);
              }
            } catch (err) {
              _didIteratorError6 = true;
              _iteratorError6 = err;
            } finally {
              try {
                if (!_iteratorNormalCompletion6 && _iterator6.return) {
                  _iterator6.return();
                }
              } finally {
                if (_didIteratorError6) {
                  throw _iteratorError6;
                }
              }
            }
          } else {
            vals.push(val);
          }
        }
      } catch (err) {
        _didIteratorError5 = true;
        _iteratorError5 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion5 && _iterator5.return) {
            _iterator5.return();
          }
        } finally {
          if (_didIteratorError5) {
            throw _iteratorError5;
          }
        }
      }

      return vals;
    },
    getPushStatement: function getPushStatement(arg) {
      return t.expressionStatement(t.callExpression(t.memberExpression(nodes, t.identifier('push')), [arg]));
    },
    visit: function visit(node, mode, block) {
      if (typeof this['visit' + node.type] === 'function') {
        lastLine = node.line;
        return this['visit' + node.type](node, mode, block);
      } else {
        throw new Error(node.type + ' is not yet supported');
      }
    },

    // [ "JSXExpressionContainer", "ConditionalExpression", "IfStatement" ]
    visitCase: function visitCase(node, mode, block) {
      // compile the case/when into a series of conditionals then compile that down to jsx
      var base = { alternate: null };
      var parent = base;
      var defaultValue = { type: 'Block', nodes: [] };
      node.block.nodes.forEach(function (whenNode) {
        if (whenNode.expr === 'default') {
          defaultValue = whenNode.block;
        } else {
          parent.alternate = {
            type: 'Conditional',
            test: node.expr + ' === ' + whenNode.expr,
            consequent: whenNode.block,
            alternate: null
          };
          parent = parent.alternate;
        }
      });
      if (!base.alternate) {
        throw new Error('Empty case blocks are not supported');
      }
      parent.alternate = defaultValue;
      return this.visit(base.alternate, mode, block);
    },

    // [ "JSX", "Expression", "Statement" ]
    visitCode: function visitCode(node, mode, block) {
      if (node.buffer && !node.mustEscape) {
        throw new Error('Unescaped, buffered code is not supported in react-pug');
      }
      switch (mode) {
        case 'jsx':
          return this.wrapExpressionInJSX(this.visitCode(node, 'jsExpression', block));
        case 'jsExpression':
          if (node.buffer) {
            return this.parseExpression(node.val);
          }
          // TODO: hoist and rename `const` and `let` variables
          return t.doExpression(t.blockStatement([this.parseStatement(node.val), t.expressionStatement(t.nullLiteral())]));
        case 'jsStatements':
          if (node.buffer) {
            return [this.getPushStatement(this.parseExpression(node.val))];
          }
          return [this.parseStatement(node.val)];
        default:
          throw new Error('Unexpected mode "' + mode + '" should be "jsx", "jsStatement" or "jsExpression"');
      }
    },

    // [ "JSXExpressionContainer", "ConditionalExpression", "IfStatement" ]
    visitConditional: function visitConditional(node, mode, block) {
      var _this = this;

      if (node.alternate && node.alternate.type === 'Conditional') {
        node.alternate = { nodes: [node.alternate] };
      }
      var test = this.parseExpression(node.test);
      switch (mode) {
        case 'jsx':
          return t.jSXExpressionContainer(this.visitConditional(node, 'jsExpression', block));
        case 'jsExpression':
          var condConsequentBlock = this.staticBlock(block);
          var condConsequent = this.joinJsExpressions(node.consequent.nodes.map(function (node) {
            return _this.visit(node, 'jsExpression', condConsequentBlock);
          }).filter(Boolean));
          condConsequentBlock.end();
          var condAlternate = t.nullLiteral();
          if (node.alternate) {
            (function () {
              var alternateBlock = _this.staticBlock(block);
              condAlternate = _this.joinJsExpressions(node.alternate.nodes.map(function (node) {
                return _this.visit(node, 'jsExpression', alternateBlock);
              }).filter(Boolean));
              alternateBlock.end();
            })();
          }
          return t.conditionalExpression(test, condConsequent, condAlternate);
        case 'jsStatements':
          var ifConsequentBlock = this.staticBlock(block);
          var ifConsequent = t.blockStatement(this.joinJsStatements(node.consequent.nodes.map(function (node) {
            return _this.visit(node, 'jsStatements', ifConsequentBlock);
          })));
          ifConsequentBlock.end();
          var ifAlternate = null;
          if (node.alternate) {
            (function () {
              var alternateBlock = _this.staticBlock(block);
              ifAlternate = t.blockStatement(_this.joinJsStatements(node.alternate.nodes.map(function (node) {
                return _this.visit(node, 'jsStatements', alternateBlock);
              })));
              alternateBlock.end();
            })();
          }
          return [t.ifStatement(test, ifConsequent, ifAlternate)];
        default:
          throw new Error('Unexpected mode "' + mode + '" should be "jsx", "jsStatement" or "jsExpression"');
      }
    },

    // [ "JSXExpressionContainer", "DoExpression", "WhileStatement" ]
    visitEach: function visitEach(node, mode, block) {
      var _this2 = this;

      if (mode === 'jsx') {
        return this.wrapExpressionInJSX(this.visitEach(node, 'jsExpression', block));
      }
      if (mode === 'jsExpression') {
        var _variableDeclaration = t.variableDeclaration('const', [t.variableDeclarator(nodes, t.arrayExpression([]))]);
        var end = t.expressionStatement(nodes);
        return t.doExpression(t.blockStatement([_variableDeclaration].concat(_toConsumableArray(this.visitEach(node, 'jsStatements', block)), [end])));
      }
      var childBlock = this.dynamicBlock(block, node.line);
      var obj = path.scope.generateUidIdentifier('pug_arr');
      var objLength = t.memberExpression(obj, t.identifier('length'));
      var variableDeclaration = t.variableDeclaration('const', [t.variableDeclarator(obj, this.parseExpression(node.obj))]);
      var index = node.key ? t.identifier(node.key) : path.scope.generateUidIdentifier('pug_index');
      var init = t.variableDeclaration('let', [t.variableDeclarator(index, t.numericLiteral(0))]);
      var test = t.binaryExpression('<', index, objLength);
      var update = t.updateExpression('++', index);
      var loop = t.forStatement(init, test, update, t.blockStatement([t.variableDeclaration('const', [t.variableDeclarator(t.identifier(node.val), t.memberExpression(obj, index, true))])].concat(this.joinJsStatements(node.block.nodes.map(function (node) {
        return _this2.visit(node, 'jsStatements', childBlock);
      })))));
      childBlock.end();
      if (node.alternate) {
        (function () {
          var alternateBlock = _this2.staticBlock(block);
          var alternate = t.blockStatement(_this2.joinJsStatements(node.alternate.nodes.map(function (node) {
            return _this2.visit(node, 'jsStatements', alternateBlock);
          }).filter(Boolean)));
          alternateBlock.end();
          loop = t.ifStatement(objLength, loop, alternate);
        })();
      }
      return [variableDeclaration, t.ifStatement(t.callExpression(t.memberExpression(t.identifier('Array'), t.identifier('isArray')), [obj]), t.blockStatement([loop]), t.blockStatement([t.throwStatement(t.newExpression(t.identifier('TypeError'), [t.stringLiteral('Expected ' + node.obj + ' to be an array.')]))]))];
    },

    // returns ["JSXElement", "JSXElement", "PushStatement"] ]
    visitTag: function visitTag(node, mode, block) {
      var _this3 = this;

      var name = t.jSXIdentifier(node.name);
      var children = (node.code ? [this.visitCode(node.code, 'jsx', this.baseBlock())] : []).concat(node.block.nodes.map(function (node) {
        return _this3.visit(node, 'jsx', _this3.baseBlock());
      }).filter(Boolean))
      // TODO: assert that these are all valid jsx things
      ;

      if (node.attributeBlocks.length) {
        throw new Error('Attribute blocks are not yet supported in react-pug');
      }

      var classes = [];
      var attrs = node.attrs.map(function (_ref2) {
        var name = _ref2.name;
        var val = _ref2.val;
        var mustEscape = _ref2.mustEscape;

        if (/\.\.\./.test(name) && val === true) {
          return t.jSXSpreadAttribute(_this3.parseExpression(name.substr(3)));
        }
        switch (name) {
          case 'for':
            name = 'htmlFor';
            break;
          case 'maxlength':
            name = 'maxLength';
            break;
          case 'class':
            name = 'className';
            break;
        }
        var expr = _this3.parseExpression(val === true ? 'true' : val);
        if (!mustEscape && (!t.isStringLiteral(expr) || /(\<\>\&)/.test(val))) {
          throw new Error('Unescaped attributes are not supported in react-pug');
        }
        if (name === 'className') {
          classes.push(expr);
          return null;
        }
        var jsxValue = t.isStringLiteral(expr) || t.isJSXElement(expr) ? expr : t.jSXExpressionContainer(expr);
        if (/\.\.\./.test(name)) {
          throw new Error('spread attributes must not have a value');
        }
        return t.jSXAttribute(t.jSXIdentifier(name), jsxValue);
      }).filter(Boolean);
      if (classes.length) {
        var value = classes.every(function (cls) {
          return t.isStringLiteral(cls);
        }) ? t.stringLiteral(classes.map(function (cls) {
          return cls.value;
        }).join(' ')) : t.jSXExpressionContainer(t.callExpression(t.memberExpression(t.arrayExpression(classes), t.identifier('join')), [t.stringLiteral(' ')]));
        attrs.push(t.jSXAttribute(t.jSXIdentifier('className'), value));
      }
      block.handleAttributes(attrs);
      var open = t.jSXOpeningElement(name, attrs, // Array<JSXAttribute | JSXSpreadAttribute>
      children.length === 0);
      var close = children.length === 0 ? null : t.jSXClosingElement(name);
      var element = t.jSXElement(open, close, children, // ["StringLiteral","JSXExpressionContainer","JSXElement"]
      children.length === 0);
      switch (mode) {
        case 'jsx':
        case 'jsExpression':
          return element;
        case 'jsStatements':
          return [this.getPushStatement(element)];
        default:
          throw new Error('Unexpected mode "' + mode + '" should be "jsx", "jsStatement" or "jsExpression"');
      }
    },

    // [ "JSXText", "StringLiteral", "PushStatement"]
    visitText: function visitText(node, mode, block) {
      switch (mode) {
        case 'jsx':
          return t.jSXText(node.val);
        case 'jsExpression':
          return t.stringLiteral(node.val);
        case 'jsStatements':
          return [this.getPushStatement(this.visitText(node, 'jsExpression', block))];
        default:
          throw new Error('Unexpected mode "' + mode + '" should be "jsx", "jsStatement" or "jsExpression"');
      }
    },

    // [ "JSXExpressionContainer", "DoExpression", "WhileStatement" ]
    visitWhile: function visitWhile(node, mode, block) {
      var _this4 = this;

      if (mode === 'jsx') {
        return this.wrapExpressionInJSX(this.visitWhile(node, 'jsExpression', block));
      }
      if (mode === 'jsExpression') {
        var variableDeclaration = t.variableDeclaration('const', [t.variableDeclarator(nodes, t.arrayExpression([]))]);
        var end = t.expressionStatement(nodes);
        return t.doExpression(t.blockStatement([variableDeclaration].concat(_toConsumableArray(this.visitWhile(node, 'jsStatements', block)), [end])));
      }
      var childBlock = this.dynamicBlock(block, node.line);
      var test = this.parseExpression(node.test);
      var body = t.blockStatement(this.joinJsStatements(node.block.nodes.map(function (node) {
        return _this4.visit(node, 'jsStatements', childBlock);
      }).filter(Boolean)));
      childBlock.end();
      return [t.whileStatement(test, body)];
    }
  };
  return ast.nodes.map(function (node) {
    return compiler.visit(node, 'jsExpression', compiler.baseBlock());
  });
};

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _pugError = require('pug-error');

var _pugError2 = _interopRequireDefault(_pugError);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }